@@ -23,6 +23,7 @@
 #include "../config.h"
 #endif
 
+#include <getopt.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
@@ -70,8 +71,7 @@ struct _IconData
 DWord xpm_col_tbl_size;
 char *xpm_col_codes;
 
-const char *package = PACKAGE;
-
+static int verbose = 0, force = 0;
 
 void
 write_byte (FILE *file, Byte byte)
@@ -500,8 +500,9 @@ add_white (ColorMap *color_map, unsigned int clr_count)
 {
   RGBQuad *white;
 
-  printf ("Adding one color slot for white, since it was not"
-	  " previously found and is needed for transparency.\n");
+  if (verbose)
+    printf ("Adding one color slot for white, since it was not"
+	    " previously found and is needed for transparency.\n");
   white = color_map->color_list + clr_count;
   white->rgbRed = 255;
   white->rgbGreen = 255;
@@ -538,14 +539,16 @@ convert_color_map (XpmImage *icon, DWord ncolors)
 		     " not be displayed properly.\n");
 	}
 
-      printf ("The resulting .ico file will contain %u colors (%u max."
-	      " colors).\n", clr_count, (unsigned int) ncolors);
+      if (verbose)
+	printf ("The resulting .ico file will contain %u colors (%u max."
+		" colors).\n", clr_count, (unsigned int) ncolors);
     }
   else
     {
       color_map = NULL;
-      printf ("The image is encoded on 24bpp. There is thus no need"
-	      " for a color map.\n");
+      if (verbose)
+	printf ("The image is encoded on 24bpp. There is thus no need"
+		" for a color map.\n");
     }
 
   return color_map;
@@ -583,18 +586,6 @@ put_8bit_color (DrawData *draw_data,
   *ptr = val;
 }
 
-void
-set_bit (DrawData *draw_data, DWord x, DWord y)
-{
-  Byte *byte;
-  Word bit_num;
-
-  bit_num = (draw_data->height - y - 1) * draw_data->width + x;
-  byte = draw_data->data + (Word) (bit_num / 8);
-  
-  *byte |= 128 >> (bit_num % 8);
-}
-
 Byte *
 make_xor_mask (XpmImage *icon)
 {
@@ -639,23 +630,20 @@ make_xor_mask (XpmImage *icon)
 Byte *
 make_and_mask (XpmImage *icon, ColorMap *color_map)
 {
-  DWord mask_size, x, y;
+  DWord width, line_size, mask_size, x;
   DWord *uint_ptr;
-  DrawData *draw_data;
   RGBQuad color; /* for bpp >= 16 */
-  Byte *mask;
+  Byte *mask, *line;
 
-  mask_size = icon->width * icon->height / 8;
-
-  draw_data = malloc (sizeof (DrawData));
-  draw_data->data = calloc (mask_size, sizeof (Byte));
-  draw_data->width = icon->width;
-  draw_data->height = icon->height;
+  width = icon->width;
+  line_size = ((width + 31) / 32) * 4;
+  mask_size = icon->height * line_size;
 
+  mask = calloc(mask_size, sizeof(Byte));
   uint_ptr = (DWord *) icon->data;
 
-  for (y = 0; y < icon->height; y++)
-    for (x = 0; x < icon->width; x++)
+  for (line = mask + mask_size - line_size; line >= mask; line -= line_size)
+    for (x = 0; x < width; x++)
       {
 	if (icon->ncolors <= 256)
 	  color = *(color_map->color_list + *uint_ptr);
@@ -663,13 +651,10 @@ make_and_mask (XpmImage *icon, ColorMap *color_map)
 	  color = convert_color (icon->colorTable + *uint_ptr);
 
 	if (color.rgbReserved)
-	  set_bit (draw_data, x, y);
+	  line[x / 8] |= 128 >> (x % 8);
 	
 	uint_ptr++;
       }
-  
-  mask = draw_data->data;
-  free (draw_data);
 
   return mask;
 }
@@ -766,7 +751,8 @@ int save_icon (char *file_name, IconData *icon_data)
     }
   else
     {
-      printf ("Saving icon to \"%s\"\n", file_name);
+      if (verbose)
+	printf ("Saving icon to \"%s\"\n", file_name);
 
       save_header (ico_file, icon_data->header);
       save_dir_entry (ico_file, icon_data->icon_dir_entry);
@@ -811,7 +797,7 @@ int convert_icon (char *ico_file_name,
       bypp = bytes_per_pixel (icon->ncolors);     
 
       icon_data->xor_size = icon->width * icon->height * bypp;
-      icon_data->and_size = icon->width * icon->height / 8;
+      icon_data->and_size = ((icon->width+31)/32) * 4 * icon->height;
  
       icon_data->icon_dir_entry =
 	create_dir_entry (icon, icon_data, ncolors);
@@ -870,19 +856,27 @@ int test_exist (char *file_name)
 
   if (!access(file_name, F_OK))
     {
-      printf ("It happens a file named \"%s\" already exists.\n"
-	      "If you wish to overwrite it, please confirm by"
-	      " typing 'Y' or 'y': ", file_name);
-      scanf ("%1s", answer);
-
-      if (*answer == 'y'
-	  || *answer == 'Y')
+      if (force)
 	ret_code = 0;
       else
 	{
-	  fprintf (stderr, "Conversion has been aborted.\n");
-	  ret_code = 1;
+	  printf ("It happens a file named \"%s\" already exists.\n"
+		  "If you wish to overwrite it, please confirm by"
+		  " typing 'Y' or 'y': ", file_name);
+	  scanf ("%1s", answer);
+
+	  if (*answer == 'y'
+	      || *answer == 'Y')
+	    ret_code = 0;
+	  else
+	    {
+	      fprintf (stderr, "Conversion has been aborted.\n");
+	      ret_code = 1;
+	    }
 	}
+
+      if (verbose && !ret_code)
+	printf ("Overwriting %s\n", file_name);
     }
   else
     ret_code = 0;
@@ -913,13 +907,15 @@ process_icon (char *xpm_file_name, char *ico_file_name)
       icon_info = malloc (sizeof (XpmInfo));
       icon = malloc (sizeof (XpmImage));
 
-      printf ("Reading icon from xpm file \"%s\"\n", xpm_file_name);
+      if (verbose)
+	printf ("Reading icon from xpm file \"%s\"\n", xpm_file_name);
       x_ret_code = XpmReadFileToXpmImage (xpm_file_name, icon, icon_info);
 
       if (!x_ret_code)
 	{
 	  make_xpm_col_table (icon);
-	  display_icon_info (icon);
+	  if (verbose)
+	    display_icon_info (icon);
 	  ret_code = convert_icon (ico_file_name, icon_info, icon);
 	}
       else
@@ -980,58 +976,116 @@ deduce_ico_name (char *file_name)
 void
 copyright ()
 {
-  printf ("%s v. %s\n"
-	  "Copyright (C) 2001 Wolfgang Sourdeau\n\n",
-	  package, VERSION);
+  printf ("Copyright (C) 2001 Wolfgang Sourdeau\n",
+	  PACKAGE, VERSION);
 }
 
 void
 basic_usage ()
 {
   printf ("Usage: %s XPM_IN_FILE [ICO_OUT_FILE]\n",
-	  package);
+	  PACKAGE);
+  printf ("Try `%s --help' for more information.\n",
+	  PACKAGE);
 }
 
 void
-help_and_version (char *param)
+version ()
 {
-  if (!strcmp (param, "--version")
-      || !strcmp (param, "-V"))
+  printf ("%s %s\n", PACKAGE, VERSION);
+}
+
+void
+help ()
+{
+  printf ("This utility converts an xpm icon to a Windows ico file.\n");
+  printf ("%s [OPTIONS]... infile.xpm [outfile.ico]\n", PACKAGE);
+  printf ("\nOptional parameters:\n"
+	  "  -H, --help			display this help"
+	  " message\n"
+	  "  -f, --force			overwrite output"
+	  " without asking\n"
+	  "  -h, --usage			display a short usage"
+	  " information\n"
+	  "  -v, --verbose			display informations"
+	  " while processing icons\n"
+	  "  -V, --version			display version"
+	  " informations\n"
+	  "\n"
+	  "Example: " PACKAGE " ghfaxviewer-icon.xpm ghfv.ico\n\n");
+  copyright ();
+  printf ("This is free software; see the source for copying"
+	  " conditions. There is NO\n"
+	  "warranty; not even for MERCHANTABILITY or FITNESS"
+	  " FOR A PARTICULAR PURPOSE.\n\n"
+	  "Please report bugs to wolfgang@contre.com.\n");
+}
+
+int parse_args (int argc, char *argv[])
+{
+  int ret_code, gorc, stop_parsing, option_index;
+  char *outfile;
+
+  struct option options[] =
     {
-      printf ("This is free software; see the source for copying"
-	      " conditions. There is NO\n"
-	      "warranty; not even for MERCHANTABILITY or FITNESS"
-	      " FOR A PARTICULAR PURPOSE.\n");
-    }
-  else if (!strcmp (param, "--usage")
-	   || !strcasecmp (param, "-h"))
+      { "force", 0, NULL, 'f' },
+      { "help", 0, NULL, 'H' },
+      { "usage", 0, NULL, 'h' },
+      { "version", 0, NULL, 'V' },
+      { "verbose", 0, NULL, 'v' },
+      { NULL, 0, NULL, 0 },
+    };
+  
+  stop_parsing = 0;
+  ret_code = EXIT_SUCCESS;
+
+  while (!stop_parsing)
     {
-      basic_usage ();
-      printf ("Try `%s --help' for more information.\n",
-	      package);
-    }
-  else if (!strcmp (param, "--help")
-	   || !strcmp (param, "-H"))
-    {
-      printf ("This utility converts an xpm icon to a Windows ico file.\n");
-      basic_usage ();
-      printf ("\nOptional parameters:\n"
-	      "  -H, --help			display this help"
-	      " message\n"
-	      "  -h, --usage			display a short usage"
-	      " information\n"
-	      "  -V, --version			display version and"
-	      " licensing informations\n"
-	      "                         	(Please read them"
-	      " carefully)\n\n"
-	      "Example: %s ghfaxviewer-icon.xpm ghfv.ico\n",
-	      package);
-    }
-  else
-    {
-      basic_usage ();
-      printf ("Option `%s' not recognized.\n", package, param);
+      gorc = getopt_long (argc, argv, "fHhVv", options, &option_index);
+
+      switch (gorc)
+	{
+	case -1:
+	  if (argv[optind + 1])
+	    outfile = argv[optind + 1];
+	  else
+	    outfile = deduce_ico_name (argv[optind]);
+
+	  process_icon (argv[optind], outfile);
+	  stop_parsing = 1;
+	  break;
+	case 'f':
+	  force = 1;
+	  break;
+	case 'H':
+	  help ();
+	  stop_parsing = 1;
+	  break;
+	case 'h':
+	  basic_usage ();
+	  stop_parsing = 1;
+	  break;
+	case 'V':
+	  version ();
+	  stop_parsing = 1;
+	  break;
+	case 'v':
+	  verbose = 1;
+	  break;
+	case '?':
+	  basic_usage ();
+	  printf ("Option `%s' not recognized.\n", argv[optind - 1]);
+	  stop_parsing = 1;
+	  ret_code = EXIT_FAILURE;
+	  break;
+	default:
+	  printf ("strange.\n");
+	  stop_parsing = 1;
+	  ret_code = EXIT_FAILURE;       
+	}
     }
+
+  return ret_code;
 }
 
 int main (int argc, char *argv[])
@@ -1039,35 +1093,7 @@ int main (int argc, char *argv[])
   int ret_code;
   char *xpm_file_name, *ico_file_name;
 
-  copyright ();
-
-  if (argc == 2)
-    {
-      xpm_file_name = argv[1];
-
-      if (*argv[1] != '-')
-	{
-	  ico_file_name = deduce_ico_name (argv[1]);      
-	  ret_code = process_icon (xpm_file_name, ico_file_name);
-	}
-      else
-	{
-	  help_and_version (argv[1]);
-	  ret_code = EXIT_SUCCESS;
-	}
-    }
-  else if (argc == 3)
-    {
-      xpm_file_name = argv[1];
-      ico_file_name = argv[2];
-      ret_code = process_icon (xpm_file_name, ico_file_name);
-    }
-  else
-    {
-      fprintf (stderr, "%s needs a xpm filename as argument\n",
-	       PACKAGE);
-      ret_code = EXIT_FAILURE;
-    }
+  ret_code = parse_args (argc, argv);
 
   return ret_code;
 }
